# $Id: //Depot/Source/SFW/riptable/Python/core/riptable/tests/test_struct.py#14 $


import keyword
import re
import sys
import unittest
from collections import OrderedDict

# from io import StringIO
import numpy as np

import riptable as rt
from riptable import Categorical, Dataset, FastArray, Struct, arange
from riptable.rt_display import DisplayString
from riptable.rt_enum import TypeRegister
from riptable.rt_numpy import arange


class Struct_Test(unittest.TestCase):
    def test_col_ctor_01(self):
        dict1 = {_k: [_i] for _i, _k in enumerate("bac")}
        st = Struct(dict1)
        self.assertEqual(list(st.keys()), list(dict1))
        dict2 = {_k: [_i] for _i, _k in enumerate(["a", "_b", "μεαν"])}
        with self.assertRaises(ValueError):
            st = Struct(dict2)
        with self.assertRaises(TypeError):
            _ = Struct([])
        with self.assertRaises(TypeError):
            _ = Struct(None)
        with self.assertRaises(TypeError):
            _ = Struct(np.array([1]))
        st = Struct()
        self.assertEqual(st.shape, (0, 0))

    # changed invalid name behavior
    # allowing all to pass through, no methods will be lost
    # 04/05/2019
    # def test_col_ctor_02(self):
    #    tempsave = Struct.AllowAnyName
    #    Struct.AllowAnyName=False
    #    hold_WarnOnInvalidNames = Struct.WarnOnInvalidNames
    #    Struct.WarnOnInvalidNames = False
    #    with self.assertRaises(ValueError):
    #        _ = Struct({'a': [0], '0': [1]})
    #    with self.assertRaises(ValueError):
    #        _ = Struct({'a': [0], 'a-b-c': [1]})
    #    with self.assertRaises(ValueError):
    #        # should never be called this way in normal course of events!
    #        struct_instance = Struct({})
    #        struct_instance._validate_names(['a', 'b', 'a'])
    #    # We auto captialize now
    #    #for kwd in keyword.kwlist:
    #    #    with self.assertRaises(ValueError):
    #    #        _ = Struct({'a': [0], kwd: [1]})
    #    #for memb in dir(Struct):
    #    #    with self.assertRaises(ValueError):
    #    #        _ = Struct({'a': [0], memb: [1]})
    #    Struct.WarnOnInvalidNames = hold_WarnOnInvalidNames
    #    Struct.AllowAnyName = tempsave

    def test_col_ctor_02(self):
        inv_keys = ["True", "False", "None"]
        arr = arange(5)
        inv_dict = {k: arr for k in inv_keys}
        with self.assertWarns(UserWarning):
            st = Struct(inv_dict)

        self.assertTrue(bool(np.all(inv_keys == list(st))))
        for k in inv_keys:
            self.assertTrue(bool(np.all(st[k] == arr)))

    def test_col_ctor_03(self):
        tempsave = Struct.AllowAnyName
        Struct.AllowAnyName = False
        hold_WarnOnInvalidNames = Struct.WarnOnInvalidNames
        Struct.WarnOnInvalidNames = True
        with self.assertWarns(UserWarning):
            _ = Struct({"a": [0], "0": [1]})
        with self.assertWarns(UserWarning):
            _ = Struct({"a": [0], "a-b-c": [1]})
        with self.assertRaises(ValueError):  # still should fail even here!
            # should never be called this way in normal course of events!
            struct_instance = Struct({})
            struct_instance._validate_names(["a", "b", "a"])
        for kwd in keyword.kwlist:
            with self.assertWarns(UserWarning):
                _ = Struct({"a": [0], kwd: [1]})
        for memb in dir(Struct):
            with self.assertWarns(UserWarning):
                # We expect really stupid errors resulting from not avoiding invalid names.
                try:
                    _ = Struct({"a": [0], memb: [1]})
                except (TypeError, AttributeError):
                    pass
        Struct.WarnOnInvalidNames = hold_WarnOnInvalidNames
        Struct.AllowAnyName = tempsave

    def test_col_ctor_04(self):
        hold_UseFastArray = Struct.UseFastArray
        arr = np.ones(1000).reshape(5, -1).T
        self.assertFalse(arr.flags.c_contiguous)
        Struct.set_fast_array(True)
        st1 = Struct({"a": arr})
        self.assertFalse(st1.a.flags.c_contiguous)
        Struct.set_fast_array(False)
        st2 = Struct({"a": arr})
        self.assertFalse(st2.a.flags.c_contiguous)
        Struct.set_fast_array(hold_UseFastArray)

    def test_lock(self):
        st1 = Struct({_k: [_i] for _i, _k in enumerate(["a", "b", "c"])})
        st1._lock()
        with self.assertRaises(AttributeError):
            st1.a = 1
        with self.assertRaises(AttributeError):
            st1.d = 1
        with self.assertRaises(AttributeError):
            st1["a"] = 1
        with self.assertRaises(AttributeError):
            st1["d"] = 1
        with self.assertRaises(AttributeError):
            del st1.b
        with self.assertRaises(AttributeError):
            st1.col_remove("c")
        with self.assertRaises(AttributeError):
            _ = st1.col_pop("c")
        with self.assertRaises(AttributeError):
            _ = st1.col_rename("c", "C")
        with self.assertRaises(AttributeError):
            _ = st1.col_map({})
        with self.assertRaises(AttributeError):
            _ = st1.col_move_to_back("c")
        with self.assertRaises(AttributeError):
            _ = st1.col_move_to_front("c")
        st1._unlock()
        st1.a = 1
        st1.d = 1
        del st1.b
        self.assertEqual(list(st1.keys()), ["a", "c", "d"])
        self.assertEqual(st1.a, 1)
        self.assertEqual(st1.d, 1)

    def test_context_interface(self):
        with self.assertRaises(AttributeError):
            with Struct({_k: [_i] for _i, _k in enumerate(["a", "b", "c"])}) as st1:
                self.assertIsInstance(st1, Struct)
                self.assertEqual(list(st1.keys()), ["a", "b", "c"])
                st1._lock()
                st1.a = "cannot modify when locked"
        self.assertEqual(st1.get_ncols(), 0)
        self.assertFalse(st1.is_locked())

    def test_basic_interface(self):
        cols = ["a", "b", "c", "μεαν"]
        dict1 = {_k: [_i] for _i, _k in enumerate(cols)}
        st1 = Struct(dict1)
        self.assertEqual(list(st1.keys()), cols)
        self.assertEqual(list(st1.keys()), cols)
        self.assertEqual(list(st1), cols)
        for _idx, (_k, _v) in enumerate(st1.items()):
            self.assertEqual(st1[_k], dict1[_k])
            self.assertEqual(cols[_idx], _k)
            self.assertEqual(st1[_idx], dict1[_k])
            self.assertEqual(getattr(st1, _k), dict1[_k])
        self.assertEqual(list(st1.keys()), cols)
        self.assertEqual(list(st1), cols)
        self.assertEqual(list(reversed(st1)), list(reversed(cols)))
        self.assertEqual(st1.get_nrows(), 0)
        self.assertEqual(st1.shape, (0, st1.get_ncols()))
        self.assertEqual(len(st1), st1.get_ncols())
        st1["a"] = -999
        self.assertEqual(st1.a, -999)
        st1["newcol"] = -999
        self.assertEqual(st1.newcol, -999)

    def test_col_indexing(self):
        cols = ["b", "a", "c", "μεαν"]
        dict1 = {_k: [_i] for _i, _k in enumerate(cols)}
        st = Struct(dict1)
        self.assertEqual(list(st.keys()), cols)
        self.assertEqual(st["b"], [0])
        self.assertEqual(st[1], [1])
        self.assertEqual(list(st[1:3]), cols[1:3])
        self.assertEqual(list(st[["b", "c"]]), ["b", "c"])
        self.assertEqual(list(st[[True, False, True, False]]), ["b", "c"])
        with self.assertRaises(IndexError):
            _ = st[[True, False, True]]
        with self.assertRaises(IndexError):
            _ = st[[True, False, True, True, True]]
        self.assertEqual(st.b, [0])
        self.assertEqual(st.a, [1])
        self.assertEqual(st.c, [2])
        self.assertEqual(st.μεαν, [3])
        # check shallow copy behavior
        st.d = "stable"
        st1 = st[["a", "b", "d"]]
        st1.a[0] = "diff1"
        st1.b = "diff2"
        st1.d = "diff3"
        self.assertEqual(st1.tolist(), [["diff1"], "diff2", "diff3"])
        self.assertEqual(st.tolist(), [[0], ["diff1"], [2], [3], "stable"])
        with self.assertRaises(AttributeError):
            _ = st.Q
        with self.assertRaises(IndexError):
            _ = st["Q"]
        with self.assertRaises(IndexError):
            _ = st[st.get_ncols()]
        with self.assertRaises(IndexError):
            _ = st[0, "a"]
        with self.assertRaises(IndexError):
            _ = st[["a", "a"]]
        with self.assertRaises(TypeError):
            st[[complex(1)]] = "complex is naughty"
        st1["a"] = "now can do this"
        self.assertEqual(st1.a, "now can do this")
        st1[["a", "b"]] = ["now can do this, too", "as well as this"]
        self.assertEqual(st1.a, "now can do this, too")
        self.assertEqual(st1.b, "as well as this")
        st1[["a", "b"]] = ("a", "b")
        self.assertEqual(st1.a, "a")
        self.assertEqual(st1.b, "b")
        st1["now_there"] = "and even this"
        self.assertEqual(st1.now_there, "and even this")
        with self.assertRaises(IndexError):
            st1[["a", "b"]] = "still cannot do this"
        with self.assertRaises(IndexError):
            st1[["a", "b"]] = np.array(["or", "this"])
        with self.assertRaises(IndexError):
            st1[["a", "b"]] = "AB"

    def test_col_rename(self):
        tempsave = Struct.AllowAnyName
        Struct.AllowAnyName = False
        dict1 = {_k: [_i] for _i, _k in enumerate("abc")}
        st = Struct(dict1)
        with self.assertRaises(ValueError):
            st.col_rename("b", "no-longer-b")
        orig_cols = list(st.keys())
        st.col_rename("a", "a")  # no-op
        with self.assertRaises(ValueError):
            st.col_rename("d", "c")  # cannot rename non-existent column
        self.assertEqual(list(st.keys()), orig_cols)
        with self.assertRaises(ValueError):
            st.col_rename("a", "c")
        for kwd in keyword.kwlist:
            with self.assertRaises(ValueError):
                st.col_rename("a", kwd)
        for memb in dir(Struct):
            with self.assertRaises(ValueError):
                st.col_rename("a", memb)
        st.col_rename("b", "no_longer_b")
        self.assertEqual(list(st.keys()), ["a", "no_longer_b", "c"])
        with self.assertRaises(TypeError):
            st.col_rename("a")
        Struct.AllowAnyName = tempsave

    def test_col_map_01(self):
        orig_cols = ["a", "b", "c"]
        dict1 = {_k: [_i] for _i, _k in enumerate(orig_cols)}
        st = Struct(dict1)
        with self.assertRaises(ValueError):
            st.col_map({"e": "f"})  # cannot rename non-existent column
        self.assertEqual(list(st.keys()), orig_cols)
        st.col_map({_k: "{}_{}".format(_k, _i) for _i, _k in enumerate(st)})
        dict2 = {"{}_{}".format(_k, _i): [_i] for _i, _k in enumerate(orig_cols)}
        st2 = Struct(dict2)
        self.assertEqual(list(st.keys()), list(st2.keys()))
        # dicts no longer equal
        # for _k in st:
        #    self.assertEqual(st.__dict__[_k], st2.__dict__[_k])
        st = Struct(dict1)
        self.assertEqual(list(st.keys()), orig_cols)
        with self.assertRaises(ValueError):
            st.col_map({"a": "e", "b": "d", "c": "d"})
            st.col_map({"q": "z"})  # cannot rename non-existent column
            st.col_map({"q": "a"})  # cannot rename non-existent column
        with self.assertRaises(TypeError):
            st.col_map({"a": "e", "b": "d", "c": 1})
        with self.assertRaises(TypeError):
            st.col_map({1: "e", "b": "d", "c": 1})
        with self.assertRaises(TypeError):
            st.col_map("a")
        self.assertEqual(list(st.keys()), orig_cols)
        st.col_map({"a": "f", "b": "e", "c": "d"})
        self.assertEqual(list(st.keys()), ["f", "e", "d"])
        self.assertEqual(st.f, [0])
        self.assertEqual(st.e, [1])
        self.assertEqual(st.d, [2])

    def test_col_map_02(self):
        st = Struct({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghijklmnopqrst")})
        cmap = {
            "a": "x",  # 1-transition
            "b": "c",
            "c": "b",  # 2-cycle
            "f": "f",  # 1-cycle
            "g": "h",
            "h": "j",
            "j": "k",
            "k": "g",  # 4-cycle
            "m": "n",
            "n": "o",
            "o": "y",  # 3-transition
            "p": "q",
            "q": "z",
            "z": "r",
        }  # false 3-cycle, really a 2-transition
        st.col_map(cmap)
        for _k in "abcdefghijklmnopqrst":
            self.assertEqual(st[cmap.get(_k, _k)][1], (ord(_k) - ord("a")) * 10 + 1)

    def test_col_map_03(self):
        # same as test_col_map_02 but w/ transitions in scrambled order
        st = Struct({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghijklmnopqrst")})
        cmap = {
            "a": "x",  # 1-transition
            "b": "c",
            "c": "b",  # 2-cycle
            "f": "f",  # 1-cycle
            "g": "h",
            "k": "g",
            "h": "j",
            "j": "k",  # 4-cycle
            "n": "o",
            "m": "n",
            "o": "y",  # 3-transition
            "q": "z",
            "p": "q",
            "z": "r",
        }  # false 3-cycle, really a 2-transition
        st.col_map(cmap)
        for _k in "abcdefghijklmnopqrst":
            self.assertEqual(st[cmap.get(_k, _k)][1], (ord(_k) - ord("a")) * 10 + 1)

    def test_col_swap(self):
        orig_cols = ["a", "b", "c", "d"]
        dict1 = {_k: [_i] for _i, _k in enumerate(orig_cols)}
        st = Struct(dict1)
        for _i, _k in enumerate(orig_cols):
            self.assertEqual(st[_k][0], _i)
        st.a, st.b = st.b, st.a
        self.assertEqual(st["a"][0], 1)
        self.assertEqual(st["b"][0], 0)
        st.col_swap(["a", "b"], ["b", "a"])
        for _i, _k in enumerate(orig_cols):
            self.assertEqual(st[_k][0], _i)
        st.col_swap(["a", "b", "c"], ["c", "b", "a"])
        self.assertEqual(st["a"][0], 2)
        self.assertEqual(st["b"][0], 1)
        self.assertEqual(st["c"][0], 0)
        self.assertEqual(list(st.keys()), orig_cols)
        st = Struct(dict1)
        st.col_swap(["a", "b", "c", "d"], ["c", "b", "a", "d"])
        self.assertEqual(st["a"][0], 2)
        self.assertEqual(st["b"][0], 1)
        self.assertEqual(st["c"][0], 0)
        self.assertEqual(st["d"][0], 3)
        self.assertEqual(list(st.keys()), orig_cols)

    def test_col_moves(self):
        st = Struct({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghijklmnop")})

        st.col_move_to_front(1)
        self.assertEqual(list(st), list("bacdefghijklmnop"))
        st.col_move_to_front(1)

        st.col_move_to_back(14)
        self.assertEqual(list(st), list("abcdefghijklmnpo"))
        st.col_move_to_back(14)

        with self.assertRaises(ValueError):
            st.col_move_to_front(arange(20))

        st.col_move_to_back(list("dgh"))
        self.assertEqual(list(st), list("abcefijklmnopdgh"))
        st.col_move_to_front(list("gpha"))
        self.assertEqual(list(st), list("gphabcefijklmnod"))
        st.col_move(list("cim"), list("hfo"))
        self.assertEqual(list(st), list("cimgpabejklndhfo"))
        st.col_move_to_front({"g": 1})
        st.col_move_to_front("h")
        with self.assertWarns(UserWarning):
            st.col_move_to_front("q")
        self.assertEqual(list(st), list("hgcimpabejklndfo"))
        st.col_move_to_back({"g": 1})
        st.col_move_to_back("h")
        with self.assertWarns(UserWarning):
            st.col_move_to_back("q")
        self.assertEqual(list(st), list("cimpabejklndfogh"))

    def test_col_remove(self):
        cols = ["aa", "b", "c", "μεαν"]
        dict1 = {_k: [_i] for _i, _k in enumerate(cols)}
        st1 = Struct(dict1)
        self.assertEqual(list(st1.keys()), cols)
        st1.col_remove("aa")
        self.assertEqual(list(st1.keys()), cols[1:])
        with self.assertRaises(IndexError):
            st1.col_remove("aa")
        st1.col_remove(["b", "c"])
        self.assertEqual(list(st1.keys()), cols[3:])
        with self.assertRaises(IndexError):
            st1.col_remove(["μεαν", "b", "c"])
        with self.assertRaises(TypeError):
            st1.col_remove(2)
        self.assertEqual(list(st1.keys()), cols[3:])
        st1.col_remove(["μεαν"])
        self.assertEqual(list(st1.keys()), [])
        st1.col_remove([])
        self.assertEqual(list(st1.keys()), [])

    def test_col_filter(self):
        x = arange(3)
        ds = Dataset({"a": x, "b": x, "ca": x, "cb": x, "d": x, "e": x, "ear": x})
        self.assertEqual(ds.col_filter(["a", "b"], "c", "e.r").keys(), ["a", "b", "ca", "cb", "ear"])

    def test_col_pop(self):
        cols = ["aa", "b", "c", "μεαν"]
        dict1 = {_k: [_i] for _i, _k in enumerate(cols)}
        st1 = Struct(dict1)
        val = st1.col_pop("aa")
        self.assertEqual(val, [0])
        self.assertEqual(list(st1.keys()), cols[1:])
        with self.assertRaises(IndexError):
            st1.col_pop("aa")
        with self.assertRaises(IndexError):
            st1.col_pop(["aa"])
        st2 = st1.col_pop(["b", "c"])
        self.assertEqual(list(st1.keys()), cols[3:])
        self.assertEqual(list(st2.keys()), ["b", "c"])
        self.assertEqual(st2.b, [1])
        self.assertEqual(st2.c, [2])
        st3 = st1.col_pop(slice(None))
        self.assertEqual(st1.get_ncols(), 0)
        self.assertEqual(st3.get_ncols(), 1)
        self.assertEqual(st3.μεαν, [3])

    def test_col_str_match(self):
        cols = ["apple", "orange", "kumquat", "koala", "μεαν", "two_kumquats_or_more"]
        dict1 = {_k: [_i] for _i, _k in enumerate(cols)}
        st1 = Struct(dict1)
        self.assertEqual(st1.col_str_match(r".*kum.*t").shape, (st1.get_ncols(),))
        self.assertEqual(st1.col_str_match(r".*kum.*t").sum(), 2)
        self.assertEqual(
            list(st1[st1.col_str_match(r".*kum.*t")]),
            ["kumquat", "two_kumquats_or_more"],
        )
        self.assertEqual(
            list(st1[st1.col_str_match(r".*KuM.*t", flags=re.IGNORECASE)]),
            ["kumquat", "two_kumquats_or_more"],
        )

    @staticmethod
    def _sum_dict(dd):
        return {_k: sum(_v) for _k, _v in dd.items()}

    def test_as_dictionary(self):
        od = {_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghij")}
        st = Struct(od)
        dt1 = st.asdict()
        dt2 = st.asdict(copy=True)
        self.assertEqual(self._sum_dict(dt1), self._sum_dict(od))
        self.assertIsInstance(dt1, dict)
        self.assertEqual(list(st.asdict(sublist=["a", "c", "d"]).keys()), ["a", "c", "d"])
        with self.assertRaises(AttributeError):
            _ = st.asdict(sublist=["a", "c", "q"]).keys()
        # test for copy/no-copy?
        lst10 = list(range(10))
        self.assertEqual(st.a, lst10)
        self.assertEqual(dt1["a"], lst10)
        self.assertEqual(dt2["a"], lst10)
        st.a[5] = -1
        self.assertEqual(st.a[5], -1)
        self.assertEqual(dt1["a"][5], -1)
        self.assertEqual(dt2["a"], lst10)
        dt2["b"][3] = -2
        self.assertEqual(st.b[3], 13)
        self.assertEqual(dt1["b"][3], 13)

    def test_as_ordered_dictionary(self):
        od = {_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghij")}
        st = Struct(od)
        dt1 = st.as_ordered_dictionary()
        self.assertEqual(self._sum_dict(dt1), self._sum_dict(od))
        self.assertIsInstance(dt1, OrderedDict)
        self.assertEqual(
            list(st.as_ordered_dictionary(sublist=["a", "c", "d"]).keys()),
            ["a", "c", "d"],
        )
        with self.assertRaises(ValueError):
            _ = st.as_ordered_dictionary(sublist=["a", "c", "q"]).keys()

    def test_no_bool(self):
        with self.assertRaises(ValueError):
            _ = bool(Struct({"A": 5}))
        with self.assertRaises(ValueError):
            _ = bool(Struct({"A": [5.5], "B": "fish"}))
        with self.assertRaises(ValueError):
            _ = bool(Struct({"A": True}))
        with self.assertRaises(ValueError):
            _ = bool(Struct({"A": True, "B": False}))

    def test_tolist(self):
        data = {"a": 5, "b": 5.6, "c": "fish", "d": "μεαν", "e": True, "f": {"A": None}}
        st = Struct(data)
        self.assertEqual(st.tolist(), list(data.values()))

    def test_any(self):
        self.assertTrue(Struct({"A": True, "B": True}).any())
        self.assertTrue(Struct({"A": True, "B": False}).any())
        self.assertFalse(Struct({"A": False, "B": False}).any())
        self.assertTrue(Struct({"A": True}).any())
        self.assertFalse(Struct({"A": False}).any())
        self.assertTrue(Struct({"A": [5.5], "B": "fish"}).any())
        self.assertTrue(Struct({"A": np.nan, "B": False}).any())
        self.assertTrue(Struct({"A": ["A", "B", ""], "B": False}).any())
        self.assertTrue(Struct({"A": np.array(["A", "B", ""]), "B": False}).any())
        self.assertFalse(Struct({}).any())  # there does not exist one which is true

    def test_all(self):
        self.assertTrue(Struct({"A": True, "B": True}).all())
        self.assertFalse(Struct({"A": True, "B": False}).all())
        self.assertFalse(Struct({"A": False, "B": False}).all())
        self.assertTrue(Struct({"A": True}).all())
        self.assertFalse(Struct({"A": False}).all())
        self.assertTrue(Struct({"A": [5.5], "B": "fish"}).all())
        self.assertFalse(Struct({"A": np.nan, "B": False}).all())
        self.assertFalse(Struct({"A": ["A", "B", ""], "B": False}).all())
        self.assertFalse(Struct({"A": np.array(["A", "B", ""]), "B": False}).all())
        self.assertTrue(Struct({}).all())  # all that exist are true

    def test_comparison_01(self):
        st1 = Struct({"A": [5.5], "B": "fish"})
        st2 = Struct({"A": [5.5], "B": "fish", "C": [380, 20]})
        st3 = Struct({"B": "fish", "C": [380, 20]})
        st4 = Struct({"A": [5.5], "B": np.ones(4), "C": [380, 20]})
        self.assertEqual((st1 == st1).asdict(), {"A": True, "B": True})
        self.assertEqual((st1 == st2).asdict(), {"A": True, "B": True, "C": False})
        self.assertEqual((st2 == st1).asdict(), {"A": True, "B": True, "C": False})
        self.assertEqual(list((st2 == st3).keys()), ["A", "B", "C"])
        self.assertEqual(list((st3 == st2).keys()), ["B", "C", "A"])
        self.assertEqual((st4 == st4).asdict(), {"A": True, "B": True, "C": True})
        for comp in "eq ne lt le gt ge".split():
            self.assertIsInstance((getattr(st1, f"__{comp}__")(st2)).all(), bool)

    def test_comparison_02(self):
        st1 = Struct({"A": [5.5], "B": 20})
        st2 = Struct({"A": [5.5], "B": "20"})
        st3 = Struct({"A": [6.5], "B": 30})
        self.assertEqual((st1 < st3).asdict(), {"A": True, "B": True})
        with self.assertRaises(NotImplementedError):
            _ = st2 < st3
        with self.assertRaises(TypeError):
            _ = st2 < []

    def test_concat(self):
        st1 = Struct(
            {
                "ds": TypeRegister.Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)}),
                "arr": arange(5),
                "cat": TypeRegister.Categorical(["a", "a", "b", "c", "a"]),
            }
        )
        st2 = Struct(
            {
                "ds": TypeRegister.Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)}),
                "arr": arange(5),
                "cat": TypeRegister.Categorical(["a", "a", "b", "c", "a"]),
            }
        )

        result = Struct.concat_structs([st1, st2])
        self.assertTrue(isinstance(result.ds, TypeRegister.Dataset))
        self.assertTrue(isinstance(result.arr, TypeRegister.FastArray))
        self.assertTrue(isinstance(result.cat, TypeRegister.Categorical))

        correct_arr = np.hstack([st1.arr, st2.arr])
        self.assertTrue(bool(np.all(correct_arr == result.arr)))

        correct_cat = np.array(["a", "a", "b", "c", "a", "a", "a", "b", "c", "a"])
        self.assertTrue(bool(np.all(correct_cat == result.cat)))

        for c in st1.ds:
            correct = np.hstack([st1.ds[c], st2.ds[c]])
            self.assertTrue(bool(np.all(correct == result.ds[c])))

        order = list(st1.keys())
        result = list(result.keys())
        self.assertTrue(bool(np.all(order == result)))

    def test_categorical(self):
        ds = Dataset({"A": ["a", "b"], "B": ["c", "d"], "C": ["a", "μεαν"]})
        # to_categorical called the categorical constructor
        # removed, because the user can simply say c=Categorical(ds.A)
        # self.assertIsInstance(Dataset.to_categorical(ds.A), Categorical)
        # self.assertNotIsInstance(ds.A, Categorical)
        ds.make_categoricals("A")
        self.assertIsInstance(ds.A, Categorical)
        ds = Dataset({"A": ["a", "b"], "B": ["c", "d"], "C": ["a", "μεαν"]})
        # 9/28/2018 SJK: dataset no longer flips unicode to categorical
        # self.assertIsInstance(ds.C, Categorical)
        ds.make_categoricals(["A", "C"])
        self.assertIsInstance(ds.A, Categorical)
        self.assertIsInstance(ds.C, Categorical)
        # BUG: categorical is not retaining its base index for true unicode...
        self.assertEqual(ds.C[1], "μεαν")
        ds = Dataset({"A": ["a", "b"], "B": ["c", "d"], "C": ["a", "μεαν"], "D": [1, 2]})
        types1 = {"A": FastArray, "B": FastArray, "C": FastArray, "D": FastArray}
        for _k in ds:
            self.assertIsInstance(ds[_k], types1[_k])
        ds.make_categoricals()
        types2 = {"A": Categorical, "B": Categorical, "C": Categorical, "D": FastArray}
        for _k in ds:
            self.assertIsInstance(ds[_k], types2[_k])

        ds1 = Dataset(
            {
                "a": np.random.choice(["a", "b", "c"], 5),
                "b": np.random.choice(["a", "b", "c"], 5),
            }
        )
        ds2 = Dataset(
            {
                "a": np.random.choice(["a", "b", "c"], 5),
                "b": np.random.choice(["a", "b", "c"], 5),
            }
        )
        ds3 = Dataset(
            {
                "a": np.random.choice(["a", "b", "c"], 5),
                "b": np.random.choice(["a", "b", "c"], 5),
            }
        )
        arr = np.random.choice(["a", "b", "c"], 5)
        st = Struct({"ds1": ds1, "ds2": ds2, "ds3": ds3, "arr": arr})
        st.make_categoricals()
        for ds in [st.ds1, st.ds2, st.ds3]:
            for col in ds:
                self.assertTrue(isinstance(ds[col], Categorical))
        self.assertTrue(isinstance(st.arr, Categorical))

    def _test_output(self):
        if sys.platform != "win32":
            intt_name = b"int64"
        else:
            intt_name = b"int32"

        class Dataset1(Struct):
            pass  # dummy for testing, mimics behavior of real Dataset

        st = Struct(
            {
                "a": Dataset1({"A": range(10), "B": range(10, 20)}),
                "b": Struct({"C": 0, "D": 1, "E": 2}),
                "c": FastArray(np.arange(5)),
                "d": np.arange(5, 10),
                "e": ["abc", "def", "ghi"],
                "f": {"q": 1, "r": 2},
                "g": 3.14,
                "h": 84,
                "i": None,
                "j": slice(None),
            }
        )
        headers, spec = st.get_table_data()
        self.assertEqual(len(headers), 1)
        self.assertEqual([hd.col_name for hd in headers[0]], ["Name", "Type", "Rows", "0", "1", "2"])
        self.assertEqual(
            [_r.tolist() for _r in spec],
            [
                [b"a", b"b", b"c", b"d", b"e", b"f", b"g", b"h", b"i", b"j"],
                [
                    b"Dataset1",
                    b"Struct",
                    intt_name,
                    intt_name,
                    b"list",
                    b"dict",
                    b"float",
                    b"int",
                    b"NoneType",
                    b"slice",
                ],
                [b"2", b"3", b"5", b"5", b"3", b"2", b"0", b"0", b"0", b"0"],
                [b"A", b"", b"0", b"5", b"", b"", b"3.14", b"84", b"", b""],
                [b"B", b"", b"1", b"6", b"", b"", b"", b"", b"", b""],
                [b"", b"", b"2", b"7", b"", b"", b"", b"", b"", b""],
            ],
        )
        self.assertEqual(
            str(st),
            f"""#   Name   Type       Rows   0      1   2
-   ----   --------   ----   ----   -   -
0   a      Dataset1   2      A      B
1   b      Struct     3
2   c      {intt_name.decode()}      5      0      1   2
3   d      {intt_name.decode()}      5      5      6   7
4   e      list       3
5   f      dict       2
6   g      float      0      3.14
7   h      int        0      84
8   i      NoneType   0
9   j      slice      0                  """,
        )
        self.assertEqual(Struct._sizeof_fmt(128), "128.0 B")
        tsize = 1280
        for unit in ["K", "M", "G", "T", "P", "E", "Z", "Y"]:
            self.assertEqual(Struct._sizeof_fmt(tsize), f"1.2 {unit}B")
            tsize *= 1024
        self.assertEqual(st._last_row_stats(), "[10 columns]")

    def test_total_sizes(self):
        st = Struct(
            {
                "a": Dataset(
                    {
                        # 10x int32 => 40B
                        "A": range(10),
                        # 10x int32 => 40B
                        "B": range(10, 20),
                    }
                ),
                "b": Struct(
                    {
                        # 1x int32 => 4B
                        "C": 0,
                        # 1x int32 => 4B
                        "D": 1,
                        # 1x int32 => 4B
                        "E": 2,
                    }
                ),
                # 5x int32 => 20B
                "c": FastArray(np.arange(5)),
                # 5x int32 => 20B
                "d": np.arange(5, 10),
                # ???
                "e": ["abc", "def", "ghi"],
                "f": {
                    # 1x int32 => 4B
                    "q": 1,
                    # 1x int32 => 4B
                    "r": 2,
                },
                # 1x float64 => 8B
                "g": 3.14,
                # 1x int32 => 4B
                "h": 84,
                # ???
                "i": None,
                # ???
                "j": slice(None),
            }
        )

        # Create some duplicated/aliased data within the struct.
        st.z = st.c

        # Calculate the sizes of the Struct's data in bytes.
        (physical, logical) = st.total_sizes

        # For now, we only check that the logical size is larger than the physical size
        # (due to the presence of aliased array(s) somewhere within the Struct).
        # TODO: Strengthen this test by checking the actual computed sizes to make sure they're correct.
        self.assertLess(physical, logical, "The physical size is not less than the logical size.")

    def test_total_sizes_with_categorical(self):
        st = Struct({"c": Categorical(["aa", "bb", "cc", "dd"])})
        st.d = st.c
        (physical, logical) = st.total_sizes
        self.assertEqual(physical, logical // 2)
        self.assertGreaterEqual(physical, np.asarray(st.c).nbytes + st.c.category_array.nbytes)

    def test_tree(self):
        """sanity check that .tree() at least returns something, even for empty Struct"""
        s = Struct()
        self.assertIsInstance(s.tree(), DisplayString)
        s["foo"] = Dataset({"bar": arange(5)})
        self.assertIsInstance(s.tree(), DisplayString)

    def test_key_search(self):
        st = Struct()
        st.column1 = rt.arange(10)
        st.Apple = rt.arange(10)
        myds = rt.Dataset()
        myds.animal = rt.arange(10)
        myds.ale = rt.arange(10)
        st.opp = myds
        mystruct = rt.Struct()
        mystruct.allen = rt.arange(10)
        substruct = rt.Struct()
        myds2 = rt.Dataset()
        myds2.oval = rt.arange(10)
        myds2.elena = rt.arange(10)
        myds2.Valence = rt.arange(10)
        substruct.myds2 = myds2
        mystruct.substruct = substruct
        st.Alex = mystruct
        self.assertEqual(
            st.key_search("a.*e"), ["Apple", "Alex", "opp.ale", "Alex.allen", "Alex.substruct.myds2.Valence"]
        )


if __name__ == "__main__":
    tester = unittest.main()



__all__ = ["GetNanoTime", "GetTSC", "tic", "toc", "ticx", "tocx", "tt", "ttx", "ticp", "tocp", "ticf", "tocf", "utcnow"]

"""
Timing and profiling functionality
"""

import logging
from typing import TYPE_CHECKING, Optional

import numpy as np
import riptide_cpp as rc

from riptable.rt_enum import TypeRegister

if TYPE_CHECKING:
    from .rt_datetime import DateTimeNano


def GetNanoTime() -> int:
    """
    Returns: a long integer in unix epoch nanoseconds
    Note: this function is written as fast as possible for both Windows and Linux
    """
    return rc.GetNanoTime()


def GetTSC() -> int:
    """
    Returns: a long integer from the CPUs's current time stamp counter

    time stamp counter (TSC) are based on the CPUs's clock cycle, which is often above 1GHz
    thus GetTSC return values are guaranteed to be both unique and subsample below 1 nanosecond

    Note: this function is written as fast as possible for both Windows and Linux
    """
    return rc.GetTSC()


def utcnow(count: int = 1) -> "DateTimeNano":
    """
    Call `GetNanoTime` one or more times and return the timestamps in a :class:`~rt.rt_datetime.DateTimeNano` array.

    Parameters
    ----------
    count : int, default to 1
        The number of timestamp samples to collect.

    Returns
    -------
    DateTimeNano
        A DateTimeNano array containing the sampled timestamps (representing the current time in UTC nanoseconds).

    Examples
    --------
    >>> import riptable as rt
    >>> rt.utcnow()
    DateTimeNano([20190215 11:29:44.022382600])
    >>> rt.utcnow()._fa
    FastArray([1550248297734812800], dtype=int64)

    To make an array containing multiple timestamps:

    >>> len(rt.utcnow(1_000_000))
    1000000

    See Also
    --------
    GetNanoTime()
    datetime.datetime.utcnow()
    """
    if count == 1:
        return TypeRegister.DateTimeNano([rc.GetNanoTime()], from_tz="GMT")
    else:
        x = [rc.GetNanoTime() for i in range(count)]
        return TypeRegister.DateTimeNano(x, from_tz="GMT")


# Timing code below
def tic():
    """
    Call tic() followed by code followed by toc() to time a routine in nanoseconds.

    See Also
    --------
    toc, ticx, ticp, ticf
    """
    global TicStartTime
    TicStartTime = GetNanoTime()


def toc(logger: Optional[logging.Logger] = None) -> None:
    """
    Call tic() followed by code followed by toc() to time a routine in nanoseconds.

    Parameters
    ----------
    logger : logging.Logger, optional
        An optionally-specified logger where the collected timing information is recorded.
        If not specified (the default), the timing information is written to stdout.

    See Also
    --------
    toc, ticx, ticp, ticf
    """
    global TicStartTime
    global TocEndTime
    TocEndTime = GetNanoTime()

    delta_ns = TocEndTime - TicStartTime

    if logger:
        logger.debug("Elapsed time (ns): %d", delta_ns)
    else:
        delta = delta_ns / 1_000_000_000.0
        deltaTime = float("{0:.6f}".format(delta))
        print("Elapsed time", deltaTime, "seconds.")


# even more accurate cycle counting
def ticx():
    """
    Call ticx() followed by code followed by tocx() to time a routine in TSC

    See also: toc, ticx, ticp, ticf
    """
    global TicStartTimeX
    TicStartTimeX = GetTSC()


def tocx(logger: Optional[logging.Logger] = None) -> None:
    """
    Call ticx() followed by code followed by tocx() to time a routine in TSC

    Parameters
    ----------
    logger : logging.Logger, optional
        An optionally-specified logger where the collected timing information is recorded.
        If not specified (the default), the timing information is written to stdout.

    See also: toc, ticx, ticp, ticf
    """
    global TicStartTimeX
    global TocEndTimeX
    TocEndTimeX = GetTSC()

    # TODO: Need to handle TSC wraparound here
    delta_cycles = TocEndTimeX - TicStartTimeX
    if logger:
        logger.debug("Elapsed time (cycles): %d", delta_cycles)
    else:
        print("Elapsed time", delta_cycles, "cycles.")


def ticf():
    """
    Call ticf() followed by code followed by tocf() to time fastarrays

    See also: toc, ticx, ticp, ticf
    """
    FA = TypeRegister.FastArray
    FA._LCLEAR()
    FA._LON()


def tocf(dataset=True):
    """
    Call ticf() followed by code followed by tocf() to time fastarrays

    Parameters
    ----------
    dataset: bool, defaults to True.
        If specified, returns a Dataset. Set to False to print out instead.
    """
    FA = TypeRegister.FastArray
    FA._LOFF()
    return TypeRegister.MathLedger._LDUMP()


def ticp():
    """
    Call ticp() followed by code followed by tocp() to profile function calls

    See also: toc, ticx, ticp, ticf
    """
    import cProfile

    global pr
    pr = cProfile.Profile()
    pr.enable()


def tocp(dataset=True, logfile=None, sort="time", strip=True, stats=False, calls=False, find=None):
    """
    Call ticp() followed by code followed by tocp() to profile anything between the ticp/tocp
    tocp() may be called again to retrieve data in a different manner

    Examples
    --------
    ticp(); ds.sort_copy(by='Symbol'); tocp()._H
    ticp(); ds.sort_copy(by='Symbol'); tocp().sort_view('cumtime')._A
    ticp(); ds.sort_copy(by='Symbol'); tocp(find='rt_fastarray.py:332')._H
    ticp(); ds.sort_copy(by='Symbol'); tocp(find='rt_fastarray.py')._H
    ticp(); ds.sort_copy(by='Symbol'); ds=tocp(calls=True); ds.gb('filepath').sum()._H
    tocp(calls=True).gb(['function','filepath'])['tottime'].sum().sort_view('tottime')._A
    ticp(); ds.sort_copy(by='Symbol'); stats=tocp(stats=True);
    ticp(); ds.sort_copy(by='Symbol'); tocp(False);
    ticp(); ds.sort_copy(by='Symbol'); tocp(False, strip=False);
    ticp(); ds.sort_copy(by='Symbol'); tocp(False, sort='cumtime');

    Parameters
    ----------
    dataset=False.  set to True to return a Dataset otherwise use pstats output
    logfile=None.   set to filename to save the Dataset in SDS format
                    NOTE: consider pickling the result when stats=True to save for later analysis
    strip=True.     set to False to return full path for the filename when dataset=False
    calls=False.    set to True to include 'callee' and 'filepath' to determine caller info
    find=None.      set to a string with 'filename:lineno' to drill into those specific calls
    sort='time'     by default when dataset=False, other options include
        "calls"     --> "call count"
        "ncalls"    --> "call count"
        "cumtime"   --> "cumulative time"
        "cumulative"--> "cumulative time"
        "file"      --> "file name"
        "filename"  --> "file name"
        "line"      --> "line number"
        "module"    --> "file name"
        "name"      --> "function name"
        "nfl"       --> "name/file/line"
        "pcalls"    --> "primitive call count"
        "stdname"   --> "standard name"
        "time"      --> "internal time"
        "tottime"   --> "internal time"

    stats=False.    set to True to return all stats collected by _lsprof.c
        return all information collected by the profiler.
        Each profiler_entry is a tuple-like object with the
        following attributes:

            code          code object
            callcount     how many times this was called
            reccallcount  how many times called recursively
            totaltime     total time in this entry
            inlinetime    inline time in this entry (not in subcalls)
            calls         details of the calls

        The calls attribute is either None or a list of
        profiler_subentry objects:

            code          called code object
            callcount     how many times this is called
            reccallcount  how many times this is called recursively
            totaltime     total time spent in this call
            inlinetime    inline time (not in further subcalls)
    """
    global pr
    pr.disable()
    if stats:
        return pr.getstats()

    if dataset:
        if sort == "time":
            sort = "tottime"
        ds = snapshot_stats(pr, sort=sort, calls=calls, findfunc=find)
        if logfile is not None:
            ds.save(logfile)
        return ds
    else:
        import pstats

        if strip:
            pstats.Stats(pr).strip_dirs().sort_stats(sort).print_stats()
        else:
            pstats.Stats(pr).sort_stats(sort).print_stats()


def snapshot_stats(pr, sort="tottime", calls=True, findfunc=None):
    """
    Parameters
    ----------
    pr:

    Other Parameters
    ----------------
    sort:
    calls:
    findfunc: must be in form filename:lineno such as 'rt_fastarray:423'
              or in the form 'rt_fastarray'

    Returns
    -------
    a Dataset
    """
    import os

    if findfunc is not None:
        try:
            funcname, linenum = findfunc.split(":")
            linenum = int(linenum)
        except Exception:
            funcname = findfunc
            linenum = None

    def parse_func_info(tup):
        func_str = []

        filepath = "~"
        # module
        if tup[0] != "~":
            # parse file name
            normpath = os.path.normpath(tup[0])
            basename = os.path.basename(normpath)
            func_str.append(basename)
            filepath = normpath[: -len(basename)]

        # line_number
        if tup[1] != 0:
            func_str.append(":" + str(tup[1]))

        # func name
        if len(func_str) != 0:
            func_str.append("(" + tup[2] + ")")

        # python func
        else:
            func_str.append(tup[2])
            # to match pstats display
            func_str[0].replace("<", "{")
            func_str[0].replace(">", "}")

        return "".join(func_str), filepath

    entries = pr.getstats()
    stats = {}
    callersdicts = {}

    # def get_top_level_stats(self):
    #    for func, (cc, nc, tt, ct, callers) in self.stats.items():
    #        self.total_calls += nc
    #        self.prim_calls  += cc
    #        self.total_tt    += tt

    # call information
    # NOTE consider cython or C since this can be huge
    for entry in entries:

        code = entry.code
        if isinstance(code, str):
            func = ("~", 0, code)  # built-in functions ('~' sorts at the end)
        else:
            func = (code.co_filename, code.co_firstlineno, code.co_name)

        nc = entry.callcount  # ncalls column of pstats (before '/')s
        cc = nc - entry.reccallcount  # ncalls column of pstats (after '/')
        tt = entry.inlinetime  # tottime column of pstats
        ct = entry.totaltime  # cumtime column of pstats
        callers = {}
        callersdicts[id(entry.code)] = callers
        stats[func] = cc, nc, tt, ct, callers

    # subcall information
    for entry in entries:
        if entry.calls:
            code = entry.code
            if isinstance(code, str):
                func = ("~", 0, code)  # built-in functions ('~' sorts at the end)
            else:
                func = (code.co_filename, code.co_firstlineno, code.co_name)

            for subentry in entry.calls:
                try:
                    callers = callersdicts[id(subentry.code)]
                except KeyError:
                    continue

                nc = subentry.callcount
                cc = nc - subentry.reccallcount
                tt = subentry.inlinetime
                ct = subentry.totaltime

                if func in callers:
                    prev = callers[func]
                    nc += prev[0]
                    cc += prev[1]
                    tt += prev[2]
                    ct += prev[3]

                callers[func] = nc, cc, tt, ct

    if findfunc is not None:
        # this path is taken when user wants to drill into
        callcount = []
        ncalls = []
        tottime = []
        cumtime = []
        names = []

        for func_info, (cc, nc, tt, ct, callers) in stats.items():
            callercount = len(callers)
            if callercount > 0:
                name, filepath = parse_func_info(func_info)
                if name[0] != "<":
                    funcn, stuff = name.split(":")
                    lineno, stuff = stuff.split("(")
                    lineno = int(lineno)
                    if (linenum is None or lineno == linenum) and funcn == funcname:
                        # NOTE: not sure this is
                        for k, v in callers.items():
                            name, filepathc = parse_func_info(k)
                            cc1, nc1, tt1, ct1 = v

                            callcount.append(cc1)
                            ncalls.append(nc1)
                            tottime.append(tt1)
                            cumtime.append(ct1)
                            names.append(name)

        ds = TypeRegister.Dataset(
            {"ncalls": ncalls, "tottime": tottime, "cumtime": cumtime, "callers": callcount, "function": names}
        )

    else:
        ncalls = []
        tottime = []
        cumtime = []
        names = []
        callcount = []
        ncallers = []
        firstcaller = []
        path = []
        pathc = []

        for func_info, (cc, nc, tt, ct, callers) in stats.items():
            ncalls.append(nc)
            tottime.append(tt)
            cumtime.append(ct)
            callcount.append(cc)
            callercount = len(callers)
            ncallers.append(callercount)
            name, filepath = parse_func_info(func_info)
            names.append(name)

            # does user want more information?
            if calls:
                filepathc = "~"
                if callercount > 0:
                    firstcall = next(iter(callers))
                    firstcall, filepathc = parse_func_info(firstcall)
                    # firstcall = f'{firstcall[0]}:{firstcall[1]}({firstcall[2]})'
                else:
                    firstcall = "~"
                firstcaller.append(firstcall)
                path.append(filepath)
                pathc.append(filepathc)

        ds = TypeRegister.Dataset(
            {"ncalls": ncalls, "tottime": tottime, "cumtime": cumtime, "callers": ncallers, "function": names}
        )

        if calls:
            ds["filepath"] = path

        arr_callcount = np.asanyarray(callcount)
        ds.percallT = ds.tottime / ds.ncalls
        ds.percallC = ds.cumtime / arr_callcount

        total_tt = ds.tottime.sum()

        # check if they want information on the caller
        if calls:
            ds["callee"] = firstcaller
            ds["filepathc"] = pathc

    # NOTE: need an option for this to not SORT because that is the order
    return ds.sort_inplace(sort, ascending=False)


def tt(expression: str, loops=1, return_time=False):
    """
    tictoc time an expression in nanoseconds.  use ; to separate lines

    Args
    ----
        arg1 is a string of code to execute
        arg2 is optional and is how many loops to execute

    """
    # import __builtin__
    # __builtin__.__dict__.update(locals())
    import inspect

    frame = inspect.currentframe()

    # allow callee to use ; for new lines
    codestr = expression.replace("; ", "\n")
    codestr = codestr.replace(";", "\n")

    # compile to byte code first to eliminate compile time in calculation
    code = compile(codestr, "<string>", "exec")

    # preallocate array of floats
    aTimers = loops * [0.0]

    for i in range(loops):
        startTime = GetNanoTime()
        exec(code, frame.f_back.f_globals, frame.f_back.f_locals)
        endTime = GetNanoTime()
        aTimers[i] = (endTime - startTime) / 1000000000.0

    if loops == 1:
        deltaTime = float("{0:.6f}".format(aTimers[0]))
        if return_time:
            return deltaTime
        print("Elapsed time", deltaTime, "seconds.")
    else:
        mTime = np.median(aTimers)
        deltaTime = float("{0:.6f}".format(mTime))
        if return_time:
            return deltaTime
        print("Median", loops, "runs", deltaTime, "seconds.")


def ttx(expression: str, loops=1):
    """
    tictoc time an expression in TSC (time stamp counters).  use ; to separate lines

    Args
    ----
        arg1 is a string of code to execute
        arg2 is optional and is how many loops to execute

    """
    import inspect

    frame = inspect.currentframe()

    # allow callee to use ; for new lines
    codestr = expression.replace(";", "\n")

    # compile to byte code first to eliminate compile time in calculation
    code = compile(codestr, "<string>", "exec")

    # preallocate array of floats
    aTimers = loops * [0]

    for i in range(loops):
        startTime = GetTSC()
        exec(code, frame.f_back.f_globals, frame.f_back.f_locals)
        endTime = GetTSC()
        aTimers[i] = endTime - startTime

    if loops == 1:
        deltaTime = aTimers[0]
        print("Elapsed time", deltaTime, "cycles.")
    else:
        mTime = np.median(aTimers)
        deltaTime = mTime
        print("Median", loops, "runs", deltaTime, "cycles.")



from riptable import *


class TestCategoricalBaseIndex:
    def testbase_index_from_codes(self):
        base_0_codes = FastArray([0, 0, 1, 2, 1])
        base_1_codes = FastArray([1, 1, 2, 3, 1])
        cats = FastArray(["a", "b", "c"])

        # default this constructor to base 0 if not set
        # c = Categorical(base_0_codes, cats)
        # self.assertEqual(c.base_index, 0)
        # self.assertEqual(c._fa[0], 0)
        # self.assertEqual(c[2], 'b')

        # still use specified base index
        c = Categorical(base_0_codes, cats, base_index=0)
        assert c.base_index == 0
        assert c._fa[4] == 1
        assert c[3] == "c"

        # even if base index is 1
        c = Categorical(base_1_codes, cats, base_index=1)
        assert c.base_index == 1
        assert c._fa[3] == 3
        assert c[1] == "a"

    # ------------------------------------------------------

    def test_strings(self):
        c = Categorical(["b", "c", "a", "a", "b"])
        assert c._fa[0] == 2
        assert c.base_index == 1

        c = Categorical(["b", "c", "a", "a", "b"], base_index=0)



# -*- coding: utf-8 -*-
import argparse
import enum
import logging
import sys
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

import pkg_resources

from ..rt_enum import TypeRegister

if TYPE_CHECKING:
    from ..rt_struct import Struct

# The following import imports all the benchmarks exposed by this module which
# is run inside main. Please do not remove this import unless changes are made to main.
from ..benchmarks import *
from .runner import quick_analysis

_logger = logging.getLogger(__name__)
"""Module-level logger."""
_logger.setLevel(logging.INFO)


class ExitCode(enum.IntEnum):
    """Encode the benchmark driver valid exit codes."""

    OK = 0
    FAILED = 1
    INTERRUPTED = 2
    USAGE_ERROR = 3


class UsageFailure(Exception):
    """Raised on benchmarks invalid usage which should be handled before performing
    any further argument processing or benchmark invocations."""

    pass


def collect_benchmarks(
    benchmark_names: Optional[List[str]] = None, collect_comparisons: bool = False
) -> Mapping[str, callable]:
    """Returns the list of benchmark functions for invocation."""
    # HACK HACK HACK - Run the benchmarks via globals()
    # This requires newly added benchmarks to be exposed at the benchmarks module level
    # and imported in this global namespace.
    g = globals()

    benchmarks: Dict[str, callable] = dict()
    for k, v in g.items():
        # Only consider benchmarks specified, if any.
        if benchmark_names and k not in benchmark_names:
            continue

        # Look for functions whose name matches our convention(s) for benchmarks.
        func: Optional[callable] = None
        if k.startswith("benchmark_") or k.startswith("bench_"):
            func = v
        if collect_comparisons and k.startswith("compare_"):
            func = v

        # If this function is a benchmark, add it to the dictionary of benchmarks to run.
        if func is not None:
            benchmarks[k] = func

    return benchmarks


def _sort_dictionary(input: dict, reverse: bool = False) -> dict:
    """Rebuild a dictionary with the same keys and values but where the keys are inserted in sorted order. This can be useful for display purposes."""
    sorted_keys = sorted(input, reverse=reverse)
    return {k: input[k] for k in sorted_keys}


def _get_package_version(package_name: str) -> Optional[str]:
    package_version: Optional[str] = None

    # Attempt #1: Get the version from the installed package.
    # Use pkg_resources (from setuptools) for py37 or below; TODO: use importlib.metadata once we only target py38+
    # https://stackoverflow.com/questions/3524168/how-do-i-get-a-python-modules-version-number-through-code
    try:
        package_version = pkg_resources.get_distribution(package_name).version
    except:
        pass

    # Attempt #2: get it from the module's __version__ attribute.
    if package_version is None:
        package_module = sys.modules.get(package_name, None)
        if package_module is not None and hasattr(package_module, "__version__"):
            package_version = package_module.__version__

    return package_version


def _get_package_versions() -> Mapping[str, Any]:
    # Package versions.
    # TODO: Instead of picking specific package names, just include all loaded top-level modules? The data isn't
    #       really that large and it'll give us that much more reproducibility. The next line implements that -- just uncomment.
    #       NOTE: The hasattr() filter should be removed here -- it causes some modules like riptide_cpp not to show up in the list;
    #             removing the check pulls in a ton of other junk though too (and causes a noticeable slowdown).
    #             Refine the check so we get the modules we want (like riptide_cpp) without all the other stuff.
    # package_versions = {name: get_package_version(name) for name, m in sys.modules.items() if hasattr(m, '__version__') and '.' not in name and not name.startswith('_')}
    package_names = ["riptable", "riptide_cpp", "numpy", "pandas", "numba", "tbb"]
    package_versions = {pkg_name: _get_package_version(pkg_name) for pkg_name in package_names}
    return _sort_dictionary(package_versions)


def _capture_benchmark_metadata() -> "Struct":
    """Capture contextual metadata for a benchmark run e.g. the current date/time, riptable version, and machine information."""
    import platform

    benchmark_metadata = TypeRegister.Struct()
    benchmark_metadata["timestamp_utc_start"] = datetime.utcnow().isoformat()

    benchmark_metadata["python_version"] = platform.python_version_tuple()
    benchmark_metadata["python_implementation"] = platform.python_implementation()

    benchmark_metadata["sysname"] = platform.system()
    benchmark_metadata["nodename"] = platform.node()
    benchmark_metadata["platform"] = platform.platform()
    benchmark_metadata["platform_release"] = platform.release()
    benchmark_metadata["platform_version"] = platform.version()
    benchmark_metadata["platform_machine"] = platform.machine()
    benchmark_metadata["platform_processor"] = platform.processor()

    try:
        from cpuinfo import get_cpu_info

        benchmark_metadata["cpuinfo"] = TypeRegister.Struct(get_cpu_info())
    except:
        _logger.warning("Unable to import 'cpuinfo' package and/or read CPU information.")

    return benchmark_metadata


def _set_win_process_priority() -> Optional[bool]:
    """
    Sets the process priority class to an elevated value.

    Microbenchmarks are typically very short in duration and therefore are prone
    to noise from other code running on the same machine. Setting the process priority
    to an elevated level while running the benchmarks helps mitigate that noise
    so the results are more accurate (and also more consistent between runs).

    Returns
    -------
    success : bool, optional
        Indication of whether (or not) setting the process priority class succeeded.
        If the priority did not need to be elevated (because it was already), None is returned.
    """
    import win32api
    import win32process

    # Psuedo-handle for the current process.
    # Because this is a psuedo-handle (i.e. isn't a real handle), it doesn't need to be cleaned up.
    curr_proc_hnd = win32api.GetCurrentProcess()

    # We use the 'ABOVE_NORMAL_PRIORITY_CLASS' here, as that should be good enough to reduce general noise;
    # if necessary, we can try the 'HIGH_PRIORITY_CLASS' but that class and higher can begin to cause the system
    # to become unresponsive so we'll avoid it unless needed; or we can control it with something like a
    # 'strong_hint' bool parameter which chooses between ABOVE_NORMAL_PRIORITY_CLASS and HIGH_PRIORITY.
    target_priority_class: int = win32process.ABOVE_NORMAL_PRIORITY_CLASS

    try:
        # Get the current process priority class. If it's already equal to or higher than the class
        # we were going to set to, don't bother -- we don't want to lower it.
        current_priority_class = win32process.GetPriorityClass(curr_proc_hnd)
        if current_priority_class >= target_priority_class:
            return None

        else:
            # Try to set the priority level for the current process.
            # It can fail if the user (or process) hasn't been granted the PROCESS_SET_INFORMATION right.
            # https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-setpriorityclass
            return win32process.SetPriorityClass(curr_proc_hnd, target_priority_class)
    except:
        return False


def _setup(cli_args: argparse.Namespace) -> None:
    """Perform any pre-processing work such as configuring display settings prior to benchmark function invocations."""
    # Display all rows/columns in the resulting Datasets/Multisets.
    TypeRegister.DisplayOptions.COL_ALL = True
    TypeRegister.DisplayOptions.ROW_ALL = True

    # Set process priority to cut down on noise from other processes.
    # TODO: Implement this for Linux and macOS/Darwin too.
    #       https://linux.die.net/man/3/setpriority
    if sys.platform.startswith("win"):
        priority_was_elevated = _set_win_process_priority()
        if priority_was_elevated is None:
            pass
        elif priority_was_elevated:
            _logger.debug("Increased process priority level to reduce measurement noise.")
        else:
            _logger.warning("Unable to set the process priority class. Benchmark results may be noisier as a result.")

    else:
        _logger.warning(
            f"Setting the process priority class is not yet supported for platform '{sys.platform}'. "
            f"Benchmark results may be noisier as a result.",
            extra={"platform": sys.platform},
        )


def make_argparser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(description="Riptable benchmark runner.")

    arg_parser.add_argument(
        dest="benchmarks",
        metavar="benchmark",
        nargs="*",
        help="space delimited benchmark names to run",
    )
    arg_parser.add_argument(
        "--comparison",
        dest="comparison",
        action="store_true",
        help="additionally collect and run comparison benchmarks",
    )
    arg_parser.add_argument(
        "-c", dest="comparison", action="store_true", help="additionally collect and run comparison benchmarks"
    )
    arg_parser.add_argument("--debug", dest="debug", action="store_true", help="debug level logging")
    arg_parser.add_argument("--d", dest="debug", action="store_true", help="debug level logging")
    arg_parser.add_argument(
        "--out-file", dest="output_filename", help="The filename where benchmark results will be saved in SDS format."
    )

    # Should there be a verbose flag? does that log INFO level logging which is the current default?
    # arg_parser.add_argument('--verbose', dest='verbose', action='store_true', help='verbose mode with INFO level logging')

    return arg_parser


def main() -> ExitCode:
    # TODO: Further improvements:
    #   - use python logging facilities and indentation for more readable display results
    #   - maybe add a benchmark progress record after each benchmark step
    #     - (optionally) dump each benchmark result data to SDS and flush before proceeding to the next trial
    #     - this way we can restart the run and get partial results (i.e. results for just some of the benchmarks)
    #       if one of the benchmarks crashes e.g. due to OOM on a particular machine

    # Configure the benchmark logger to log to 'benchmarks.log' and to console.
    # Table of log record attributes: https://docs.python.org/3.9/library/logging.html#logrecord-attributes
    # Table of logging levels: https://docs.python.org/3.9/library/logging.html#levels
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    # set up root logger which writes all logs
    logging.basicConfig(filename="benchmarks.log", level=logging.INFO, format=log_format)
    formatter = logging.Formatter(log_format)

    # set up module logger that only writes benchmark related logs
    root = logging.getLogger()

    # set up console logger
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)

    _logger.addHandler(ch)

    # Parse CLI arguments.
    arg_parser = make_argparser()
    args = arg_parser.parse_args()
    _logger.info(f"main: parsed args {args}")

    if args.debug:
        root.setLevel(logging.DEBUG)
        _logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)

    # Perform other setup/configuration tasks.
    _setup(args)

    # Switch to control whether we run comparisons or not,
    # since as things are structured now they'll be doing redundant work.
    run_comparisons = args.comparison
    runnable_benchmarks: Optional[List[str]] = args.benchmarks if args.benchmarks else None

    bench_map = collect_benchmarks(benchmark_names=runnable_benchmarks, collect_comparisons=run_comparisons)
    if bench_map is None:
        raise UsageFailure(f"No benchmarks found with arguments {args}.")
    if not all(bench_map.values()):
        raise UsageFailure(f"Error collecting benchmarks.\nPassed arguments {args}\nCollected {bench_map}")

    # Struct to hold benchmark results if the option has been specified
    # to save the data.
    benchmark_results: Optional["Struct"] = TypeRegister.Struct() if args.output_filename else None

    # If saving out benchmark results, capture some additional metadata about the benchmark run,
    # e.g. the current date/time, riptable version, and machine information.
    benchmark_metadata: Optional["Struct"] = None
    if benchmark_results is not None:
        benchmark_metadata = _capture_benchmark_metadata()
        benchmark_results["meta"] = benchmark_metadata

    for name, func in bench_map.items():
        _logger.info(f"Running {name}")

        # TODO: Reset/GC recycler before each benchmark to keep peak mem usage as low as possible?

        # Run the benchmark function; it will return a Dataset containing the
        # raw results over multiple runs.
        try:
            benchmark_result = func()
        except Exception as exc:
            # Don't want the whole benchmark run to crash if there's an error when running
            # one of the benchmark functions. Log the exception and continue.
            _logger.exception(f"Exception occurred when running benchmark '{name}'.", exc_info=exc)
            continue

        # If there was no result for this benchmark (or comparison),
        # it's typically because there were no parameters for the benchmark and
        # therefore it didn't run. Log a warning about it and continue.
        if benchmark_results is None:
            _logger.warning(
                f"No results for benchmark '{name}'. This may be due to a misconfiguration of the benchmark function."
            )
            continue

        # If saving the results to SDS, save the raw/complete data for analysis --
        # we want all the data, rather than the rolled-up data we'd display to the console.
        benchmark_results[name] = benchmark_result

        # If the benchmark is one of the new-style benchmarks, we can automatically
        # perform a quick analysis to summarize the results for display to the console.
        if name.startswith("bench_") or name.startswith("compare_"):
            benchmark_result = quick_analysis(benchmark_result)

        # Write the result out to the logger (e.g. so it's written to a file and the console).
        _logger.info(f"\n{benchmark_result}")

    # If the option was specified to write the benchmark data out to a file,
    # do that now.
    if args.output_filename:
        # Capture some additional metadata about the benchmark run before saving.
        benchmark_metadata["timestamp_utc_end"] = datetime.utcnow().isoformat()
        benchmark_metadata["packages"] = _get_package_versions()

        # TODO: Capture additional metadata
        #   * numba threading engine (only available once we've executed at least one parallel numba function)
        #   * peak memory usage for this process?

        _logger.info("Writing results to: %s", args.output_filename, extra={"output_filename": args.output_filename})
        benchmark_results.save(args.output_filename, overwrite=True)
        _logger.info("Finished writing results to disk.")

    _logger.info(f"main: finished with exit code {ExitCode.OK}")
    return ExitCode.OK



from enum import IntEnum

from riptable import *

sorted_codes = FastArray(
    [
        2,
        4,
        4,
        3,
        2,
        1,
        3,
        2,
        0,
        1,
        3,
        4,
        2,
        0,
        4,
        3,
        1,
        0,
        1,
        2,
        3,
        1,
        4,
        2,
        2,
        3,
        4,
        2,
        0,
        2,
    ]
)
enum_codes = (sorted_codes + 1) * 10
str_fa = FastArray(
    [
        "c",
        "e",
        "e",
        "d",
        "c",
        "b",
        "d",
        "c",
        "a",
        "b",
        "d",
        "e",
        "c",
        "a",
        "e",
        "d",
        "b",
        "a",
        "b",
        "c",
        "d",
        "b",
        "e",
        "c",
        "c",
        "d",
        "e",
        "c",
        "a",
        "c",
    ]
)
flt_fa = FastArray(
    [
        88.97,
        50.05,
        36.16,
        63.22,
        27.33,
        17.98,
        23.35,
        52.60,
        26.82,
        78.75,
        34.20,
        49.35,
        19.50,
        29.91,
        74.96,
        94.41,
        27.95,
        56.65,
        38.08,
        16.51,
        36.81,
        80.14,
        53.64,
        49.55,
        65.82,
        85.28,
        61.68,
        72.85,
        91.71,
        61.12,
    ]
)
flt_fa.set_name("floats")
complete_unique_cats = FastArray(["a", "b", "c", "d", "e"])
unsorted_unique_cats = FastArray(["c", "e", "d", "b", "a"])
big_cats = FastArray(["string" + str(i) for i in range(2000)])
matlab_codes = (sorted_codes + 1).astype(np.float32)

sorted_float_sum = FastArray([205.09, 242.9, 454.25, 337.27, 325.84])
unsorted_float_sum = FastArray([454.25, 325.84, 337.27, 242.9, 205.09])


class str_enum(IntEnum):
    a = 10
    b = 20
    c = 30
    d = 40
    e = 50


class TestCategoricalFirstOrder:
    def test_strings(self):
        c = Categorical(str_fa)
        result = bool(np.all(c.category_array == complete_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result

        c = Categorical(str_fa, ordered=False)
        result = bool(np.all(c.category_array == unsorted_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, unsorted_float_sum))
        assert result

        c = Categorical(str_fa, complete_unique_cats)
        result = bool(np.all(c.category_array == complete_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result

        c = Categorical(str_fa, complete_unique_cats, ordered=False)
        result = bool(np.all(c.category_array == complete_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result

        c = Categorical(str_fa, unsorted_unique_cats, ordered=False, sort_gb=True)
        result = bool(np.all(c.category_array == unsorted_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result

        c = Categorical(str_fa, unsorted_unique_cats, ordered=False)
        result = bool(np.all(c.category_array == unsorted_unique_cats))
        assert result
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, unsorted_float_sum))
        assert result

    def test_multikey(self):
        c = Categorical([str_fa, sorted_codes])
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, unsorted_float_sum))
        assert result

        c = Categorical([str_fa, sorted_codes], ordered=False, sort_gb=True)
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result
        first_arr = c.category_dict["key_0"]
        result = bool(np.all(first_arr == unsorted_unique_cats))
        assert result

        c = Categorical([str_fa, sorted_codes])
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, unsorted_float_sum))
        assert result

        c = Categorical([str_fa, sorted_codes], ordered=False, sort_gb=True)
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result
        first_arr = c.category_dict["key_0"]
        result = bool(np.all(first_arr == unsorted_unique_cats))
        assert result

    def test_enum(self):
        c = Categorical(enum_codes, str_enum)
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, unsorted_float_sum))
        assert result

        c = Categorical(enum_codes, str_enum, sort_gb=True)
        float_sum = c.sum(flt_fa).floats
        result = bool(np.allclose(float_sum, sorted_float_sum))
        assert result

    def test_lex(self):
        # string sort test
        a = arange(100).astype("S")
        c1 = Cat(a, lex=True)
        c2 = Cat(a, ordered=True)
        assert np.all(c1._fa == c2._fa)
        assert np.all(c1.categories() == c2.categories())

        # make sure sorted
        srt = np.sort(c1._categories)
        assert np.all(c1._categories == srt)

        # with filter
        f = logical(arange(100) % 3)
        c1 = Cat(a, lex=True, filter=f)
        c2 = Cat(a, ordered=True, filter=f)
        assert np.all(c1._fa == c2._fa)

        # larger data
        r = np.random.randint(0, 800_000, 2_000_000)
        r1 = Cat(r, lex=True)
        r2 = Cat(r, ordered=True)
        assert np.all(r1._fa == r2._fa)

        # make sure sorted
        srt = np.sort(r1._categories)
        assert np.all(r1._categories == srt)

        f = logical(arange(len(r)) % 3)
        r1 = Cat(r, lex=True, filter=f)
        r2 = Cat(r, ordered=True, filter=f)
        assert np.all(r1._fa == r2._fa)

        # larger data multikey
        q = np.random.randint(0, 800_000, 2_000_000)
        r1 = Cat([r, q], lex=True, filter=f)
        r2 = Cat([r, q], ordered=True, filter=f)
        assert np.all(r1._fa == r2._fa)

        # multikey strings
        r = np.random.randint(0, 80, 200).astype("U")
        q = np.random.randint(0, 80, 200)
        q = q + 999_000
        q = q.astype("U")

        r1 = Cat([q, r], lex=True)
        r2 = Cat([q, r], ordered=True)
        assert np.all(r1._fa == r2._fa)

        # make sure multikey is sorted
        x = r1._categories.copy()
        p1 = x.popitem()[1]
        p2 = x.popitem()[1]

        # the dict pops in reverse order of the sort
        y = r2._categories.copy()
        y1 = y.popitem()[1]
        y2 = y.popitem()[1]
        assert np.all(p1 == y1)
        assert np.all(p2 == y2)

        p = p2 + p1
        srt = np.sort(p)
        assert np.all(p == srt)



import pandas as pd
import pytest

from riptable import *


class TestCategoricalKeywords:
    def test_invalid(self):
        c = Categorical(["b", "a", "a", "Inv", "c", "a", "b"], invalid="Inv")
        assert b"Inv" in c.category_array
        assert c.base_index == 1

        # with self.assertRaises(ValueError):
        #    c = Categorical(['b','a','a','Inv','c','a','b'], invalid='Inv', base_index=0)

        c = Categorical(["b", "a", "a", "c", "a", "b"], invalid="Inv", base_index=0)
        assert b"Inv" not in c.category_array

        with pytest.raises(ValueError):
            c = Categorical(["b", "a", "a", "Inv", "c", "a", "b"], ["a", "b", "c"])

        # 5/19/2019 invalid category must be in uniques
        with pytest.raises(ValueError):
            c = Categorical(["b", "a", "a", "Inv", "c", "a", "b"], ["a", "b", "c"], invalid="Inv")
            assert "Inv" not in c.category_array
            assert b"Inv" not in c.category_array
            assert c.base_index == 1
            assert c._fa[3] == 0
        with pytest.raises(ValueError):
            c = Categorical(
                ["b", "a", "a", "Inv", "c", "a", "b", "d", "e"],
                ["a", "b", "c"],
                invalid="Invalid",
            )
            assert "Invalid" not in c.category_array
            assert b"Invalid" not in c.category_array
            assert c.base_index == 1
            for i in [3, 7, 8]:
                assert c._fa[i] == 0

        c = Categorical([2, 1, 1, 0, 1, 2, 3], ["c", "a", "b"], invalid="Invalid")
        assert "Invalid" not in c.category_array
        assert b"Invalid" not in c.category_array
        assert c.base_index == 1
        assert c._fa[3] == 0

        # with self.assertRaises(ValueError):
        #    c = Categorical([2,1,1,0,1,2], ['c','a','b'], invalid='Invalid', base_index=0)

    def test_base_index_zero_not_allowed(self):
        with pytest.raises(ValueError):
            c = Categorical(
                ["b", "a", "a", "c", "a", "b"],
                filter=FA([True, True, False, True, True, True]),
                base_index=0,
            )
        with pytest.raises(ValueError):
            c = Categorical(
                ["b", "a", "a", "c", "a", "b"],
                ["a", "b", "c"],
                filter=FA([True, True, False, True, True, True]),
                base_index=0,
            )
        with pytest.raises(ValueError):
            c = Categorical(
                [1, 0, 0, 2, 0, 1],
                ["a", "b", "c"],
                filter=FA([True, True, False, True, True, True]),
                base_index=0,
            )
        # with self.assertRaises(ValueError):
        #    c = Categorical(['b','a','a','c','a','b'], ['a','b','c'], invalid='c', base_index=0)
        with pytest.raises(ValueError):
            pdc = pd.Categorical(["b", "a", "a", "c", "a", "b"])
            c = Categorical(pdc, base_index=0)
        with pytest.raises(ValueError):
            c = Categorical(
                [2.0, 1.0, 1.0, 3.0, 1.0, 2.0],
                ["a", "b", "c"],
                from_matlab=True,
                base_index=0,
            )
        # muted this warning to simplify categorical constructor
        # with self.assertWarns(UserWarning):
        #    c = Categorical([2,1,1,3,1,2], {'a':1, 'b':2, 'c':3}, base_index=0)



import unittest

import numpy

from riptable import Categorical, Dataset
from riptable.rt_datetime import DateTimeNano
from riptable.Utils.conversion_utils import (
    append_dataset_dict,
    dataset_as_matrix,
    dset_dict_to_list,
    numpy2d_to_dict,
    numpy_array_to_dict,
    possibly_convert_to_nanotime,
)


class Conversion_Utility_Test(unittest.TestCase):
    def test_as_matrix(self):
        error_tol = 0.00001
        ds = Dataset({"A": [1.2, 3.1, 9.6], "B": [-1.6, 2.7, 4.6]})
        X, _ = dataset_as_matrix(ds)
        self.assertIsInstance(X, numpy.ndarray)
        self.assertEqual(X.shape[0], ds.shape[0])
        self.assertEqual(X.shape[1], ds.shape[1])  # we may break this later
        self.assertTrue((numpy.abs(ds.A._np - X[:, 0]) < error_tol).all())
        self.assertTrue((numpy.abs(ds.B._np - X[:, 1]) < error_tol).all())

    def test_as_matrix_metadata(self):
        error_tol = 0.00001
        ds = Dataset(
            {
                "A": ["EXCH1", "EXCH2", "EXCH1", "EXCH3", "EXCH3"],
                "B": [-1.6, 2.7, 4.6, 5.7, 8.9],
                "C": Categorical([0, 0, 1, 0, 2], ["CPTYA", "CPTYB", "CPTYC"]),
            }
        )
        X, X_data = dataset_as_matrix(ds)
        self.assertIsInstance(X, numpy.ndarray)
        self.assertEqual(X.shape[0], ds.shape[0])
        self.assertEqual(X.shape[1], ds.shape[1])  # we may break this later
        self.assertEqual(X_data["A"]["dtype"], ds.A.dtype)
        self.assertEqual(X_data["B"]["dtype"], ds.B.dtype)
        self.assertEqual(X_data["C"]["dtype"], ds.C.dtype)
        self.assertEqual(X_data["A"]["is_categorical"], False)
        self.assertEqual(X_data["B"]["is_categorical"], False)
        self.assertEqual(X_data["C"]["is_categorical"], True)
        self.assertTrue(
            (numpy.abs(X[:, 0] - numpy.array([0.0, 1.0, 0.0, 2.0, 2.0])) < error_tol).all(), msg=f"got {X[:, 0]}"
        )
        self.assertTrue((numpy.abs(X[:, 2] - numpy.array([0, 0, 1, 0, 2])) < error_tol).all(), msg=f"got {X[:, 2]}")
        self.assertTrue(
            (X_data["A"]["category_values"][numpy.array([0, 1, 0, 2, 2])] == ds.A).all(),
            msg=f"X_data {X_data['A']['category_values'][numpy.array([0, 1, 0, 2, 2])]}\nds.A {ds.A}",
        )

    def test_as_matrix_int(self):
        error_tol = 0.00001
        ds = Dataset({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("ABCDEFGHIJKLMNOP")})
        X, _ = dataset_as_matrix(ds)
        self.assertIsInstance(X, numpy.ndarray)
        self.assertEqual(X.shape[0], ds.shape[0])
        self.assertEqual(X.shape[1], ds.shape[1])  # we may break this later
        self.assertTrue((numpy.abs(ds.A._np - X[:, 0]) < error_tol).all())
        self.assertTrue((numpy.abs(ds.B._np - X[:, 1]) < error_tol).all())

    def test_numpy_array_to_dict(self):
        arr = numpy.arange(12).reshape((3, 4)).transpose()
        cols = ["A", "C", "B"]
        dd = numpy_array_to_dict(arr, cols)
        self.assertEqual(list(dd), cols)
        self.assertTrue((dd["A"] == numpy.arange(0, 4)).all())
        self.assertTrue((dd["C"] == numpy.arange(4, 8)).all())
        self.assertTrue((dd["B"] == numpy.arange(8, 12)).all())
        arr = numpy.array([(1.0, "Q"), (-3.0, "Z")], dtype=[("x", numpy.float64), ("y", "S1")])
        dd = numpy_array_to_dict(arr)
        self.assertEqual(list(dd), ["x", "y"])
        self.assertTrue((dd["x"] == numpy.array([1.0, -3.0], dtype=numpy.float64)).all())
        self.assertTrue((dd["y"] == numpy.array(["Q", "Z"], dtype="S1")).all())

    # TODO: Remove this? -CLH
    def test_numpy2d_to_dict(self):
        arr = numpy.arange(12).reshape((3, 4)).transpose()
        cols = ["A", "C", "B"]
        dd = numpy2d_to_dict(arr, cols)
        self.assertEqual(list(dd), cols)
        self.assertTrue((dd["A"] == numpy.arange(0, 4)).all())
        self.assertTrue((dd["C"] == numpy.arange(4, 8)).all())
        self.assertTrue((dd["B"] == numpy.arange(8, 12)).all())

    def test_dset_dict_to_list(self):
        ds = Dataset({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghijklmnop")})
        ds0 = ds[:3].copy()
        ds1 = ds[6:9].copy()
        ds2 = ds[11:15].copy()
        dd = {"one": ds0, "two": ds1, "μεαν": ds2}
        with self.assertRaises(ValueError):
            _ = dset_dict_to_list(dd, "keyfield")
        dd = {"one": ds0, "two": ds1, 3: ds2}
        with self.assertRaises(ValueError):
            _ = dset_dict_to_list(dd, "keyfield")
        dd = {"one": ds0, "two": ds1, "three": ds2}
        with self.assertRaises(ValueError):
            _ = dset_dict_to_list(dd, "a")
        lst1 = dset_dict_to_list(dd, "keyfield")
        self.assertEqual(id(ds0), id(lst1[0]))
        self.assertEqual(id(ds1), id(lst1[1]))
        self.assertEqual(id(ds2), id(lst1[2]))
        self.assertEqual(list(ds0.keys()), ["a", "b", "c", "keyfield"])
        self.assertTrue((ds0.a == list(range(10))).all())
        self.assertTrue((ds0.keyfield == "one").all())
        lst2 = dset_dict_to_list(dd, "a", allow_overwrite=True)
        self.assertEqual(id(ds0), id(lst1[0]))
        self.assertEqual(list(ds0.keys()), ["a", "b", "c", "keyfield"])
        self.assertTrue((ds0.a == "one").all())
        self.assertTrue((ds0.b == list(range(10, 20))).all())
        self.assertTrue((ds0.keyfield == "one").all())

    def test_append_dataset_dict(self):
        ds = Dataset({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate("abcdefghijklmnop")})
        ds0 = ds[:3].copy()
        ds1 = ds[6:9].copy()
        ds2 = ds[11:15].copy()
        dd = {"one": ds0, "two": ds1, "three": ds2}
        ds = append_dataset_dict(dd, "keyfield")
        ucols = set()
        for _d in dd.values():
            ucols.update(_d)
        self.assertEqual(set(ds.keys()), ucols)
        self.assertEqual(ds.get_nrows(), sum(_d.get_nrows() for _d in dd.values()))
        keyfield = []
        for _d in dd.values():
            keyfield.extend(_d.keyfield)
        self.assertTrue((ds.keyfield == keyfield).all())
        self.assertTrue((ds.a[:10] == range(10)).all())
        self.assertTrue((ds.g[10:20] == range(60, 70)).all())
        self.assertTrue((ds.l[20:30] == range(110, 120)).all())

    def test_possibly_convert_to_nanotime(self):
        ns1 = numpy.int64(1528295695919153408)
        arr1 = numpy.array([ns1 + 12 * _i for _i in range(25)])
        nt1, okay1 = possibly_convert_to_nanotime(arr1)
        self.assertTrue(okay1)
        self.assertIsInstance(nt1, DateTimeNano)
        arr2 = arr1.astype(numpy.uint64)
        nt2, okay2 = possibly_convert_to_nanotime(arr2)
        self.assertFalse(okay2)
        self.assertNotIsInstance(nt2, DateTimeNano)
        self.assertTrue((nt2 == arr2).all())
        ns3 = numpy.int64(1070376029353467904)
        arr3 = numpy.array([ns3 + 12 * _i for _i in range(25)])
        nt3, okay3 = possibly_convert_to_nanotime(arr3)
        self.assertFalse(okay3)
        self.assertNotIsInstance(nt3, DateTimeNano)
        self.assertTrue((nt3 == arr3).all())


if __name__ == "__main__":
    tester = unittest.main()



import pytest

from riptable import *


def arr_eq(a, b):
    return bool(np.all(a == b))


def arr_all(a):
    return bool(np.all(a))


class TestCategoricalSetItem:
    def test_setitem_array_bool(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        b = logical(arange(5) % 2)
        c[b] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = "f"
        c.auto_add_on()
        c[b] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_array_singleint(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c[1] = "b"
        c[3] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[1] = "f"
        c.auto_add_on()
        c[1] = "f"
        c[3] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_array_index(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        b = bool_to_fancy(logical(arange(5) % 2))
        c[b] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = "f"
        c.auto_add_on()
        c[b] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_array_string(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c["a"] = "b"
        assert arr_eq(c._fa, FastArray([2, 2, 2, 3, 2]))
        with pytest.raises(ValueError):
            c["b"] = "f"
        c.auto_add_on()
        c["b"] = "f"
        assert arr_eq(c._fa, FastArray([4, 4, 4, 3, 4]))

    def test_setitem_array_slice(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c[:2] = "b"
        assert arr_eq(c._fa, FastArray([2, 2, 2, 3, 1]))
        with pytest.raises(ValueError):
            c[:2] = "f"
        c.auto_add_on()
        c[:2] = "f"
        assert arr_eq(c._fa, FastArray([4, 4, 2, 3, 1]))

    def test_setitem_mapping_bool(self):
        idx = FastArray([1, 1, 2, 3, 1])
        codes = {"a": 1, "b": 2, "c": 3}
        c = Categorical(idx, codes)
        b = logical(arange(5) % 2)
        c[b] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = "f"
        c.mapping_add(4, "f")
        c[b] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_mapping_singleint(self):
        idx = FastArray([1, 1, 2, 3, 1])
        codes = {"a": 1, "b": 2, "c": 3}
        c = Categorical(idx, codes)
        c[1] = "b"
        c[3] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[1] = "f"
        c.mapping_add(4, "f")
        c[1] = "f"
        c[3] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_mapping_index(self):
        idx = FastArray([1, 1, 2, 3, 1])
        codes = {"a": 1, "b": 2, "c": 3}
        c = Categorical(idx, codes)
        b = bool_to_fancy(logical(arange(5) % 2))
        c[b] = "b"
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = "f"
        c.mapping_add(4, "f")
        c[b] = "f"
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_mapping_string(self):
        idx = FastArray([1, 1, 2, 3, 1])
        codes = {"a": 1, "b": 2, "c": 3}
        c = Categorical(idx, codes)
        c["a"] = "b"
        assert arr_eq(c._fa, FastArray([2, 2, 2, 3, 2]))
        with pytest.raises(ValueError):
            c["b"] = "f"
        c.mapping_add(4, "f")
        c["b"] = "f"
        assert arr_eq(c._fa, FastArray([4, 4, 4, 3, 4]))

    def test_setitem_numeric_bool(self):
        c = Categorical([1.23, 1.23, 4.56, 7.89, 1.23])
        b = logical(arange(5) % 2)
        c[b] = 4.56
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = 10.1112
        c.auto_add_on()
        c[b] = 10.1112
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_numeric_singleint(self):
        c = Categorical([1.23, 1.23, 4.56, 7.89, 1.23])
        c[1] = 4.56
        c[3] = 4.56
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[1] = 10.1112
        c.auto_add_on()
        c[1] = 10.1112
        c[3] = 10.1112
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_numeric_index(self):
        c = Categorical([1.23, 1.23, 4.56, 7.89, 1.23])
        b = bool_to_fancy(logical(arange(5) % 2))
        c[b] = 4.56
        assert arr_eq(c._fa, FastArray([1, 2, 2, 2, 1]))
        with pytest.raises(ValueError):
            c[b] = 10.1112
        c.auto_add_on()
        c[b] = 10.1112
        assert arr_eq(c._fa, FastArray([1, 4, 2, 4, 1]))

    def test_setitem_numeric_string(self):
        c = Categorical([1.23, 1.23, 4.56, 7.89, 1.23])
        c[1.23] = 4.56
        assert arr_eq(c._fa, FastArray([2, 2, 2, 3, 2]))
        with pytest.raises(ValueError):
            c[4.56] = 10.1112
        c.auto_add_on()
        c[4.56] = 10.1112
        assert arr_eq(c._fa, FastArray([4, 4, 4, 3, 4]))

    def test_setitem_filter(self):
        data = Dataset({"First": Cat("A B A B A B".split())})
        data.Second = Cat("B A B A B A".split())
        data.Third = data.First.copy()
        f = data.Third == "A"
        data.First[0] = 2
        assert not arr_eq(data.First._fa, data.Third._fa)
        data.First[f] = data.Second[f]
        assert not arr_eq(data.First._fa, data.Third._fa)
        assert arr_eq(data.First._fa, FastArray([2, 2, 2, 2, 2, 2]))



import random as rand

import pandas as pd

import riptable as rt


class MultiKeyGroupBy_Test:
    ##data generation code
    def test_multkey(self):
        alpha = "Q W E R T Y U I O P A S D F G H J K L Z X C V B N M".split(" ")
        digits = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]

        sz = 4000
        numbers = [0] * sz
        keys1 = [""] * sz
        keys2 = [""] * sz

        for i in range(0, sz):
            numbers[i] = digits[rand.randint(0, 1000) % len(digits)]
            keys1[i] = alpha[rand.randint(0, 1000) % len(alpha)]
            keys2[i] = alpha[rand.randint(0, 1000) % len(alpha)]

        ary = rt.FastArray(numbers)

        data = {"k1": keys1, "k2": keys2, "beta": numbers}

        # print('SFW--------------------------------------------------------------')
        mset = rt.Dataset(data)

        # t = time.time()
        s_group = rt.GroupBy(mset, keys=["k1", "k2"]).sum()
        # print(time.time() - t, 'SFW GROUP BY ')

        # print('PANDAS--------------------------------------------------------------')
        df2 = pd.DataFrame(data)

        # t = time.time()
        p_group = df2.groupby(["k1", "k2"]).sum()
        # print(time.time() - t, 'PANDAS GROUP BY ')
        # print('compare out--------------------------------------------------------------')

        pandas = list(p_group["beta"])
        sfw = list(s_group["beta"])
        assert pandas == sfw

    def test_advanced_multikey(self):
        ##data generation code
        alpha = "Q W E R T Y U I O P A S D F G H J K L Z X C V B N M".split(" ")
        digits = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]

        sz = 200
        numb_kvs = 5  # can't be more than 26 as we are usign the contents of alpha for the column name

        # 2d array of keys/values
        vals = [[0] * sz] * numb_kvs
        keys = [[""] * sz] * numb_kvs

        # random initialization for them
        for n in range(0, numb_kvs):
            for i in range(0, sz):
                vals[n][i] = digits[rand.randint(0, 1000) % len(digits)]
                keys[n][i] = alpha[rand.randint(0, 1000) % len(alpha)]

        # create the data map
        # multi key hash for numbkeys 1:numb_kvs
        while numb_kvs > 0:
            data = {}
            for n in range(0, numb_kvs):
                data[alpha[n]] = keys[n]
                data[alpha[n + numb_kvs]] = vals[n]

            key_cols = alpha[0:numb_kvs]
            val_cols = alpha[numb_kvs : numb_kvs * 2]

            # print('SFW--------------------------------------------------------------')
            mset = rt.Dataset(data)
            # t = time.time()
            s_group = rt.GroupBy(mset, keys=key_cols).sum()
            # print(time.time() - t, 'SFW GROUP BY ')

            # print('PANDAS--------------------------------------------------------------')
            df2 = pd.DataFrame(data)
            # t = time.time()
            p_group = df2.groupby(key_cols).sum()
            # print(time.time() - t, 'PANDAS GROUP BY ')
            # print('compare out--------------------------------------------------------------')

            pandas_ = list(p_group[val_cols])
            sfw_ = list(s_group[val_cols])

            assert pandas_ == sfw_

            numb_kvs = numb_kvs - 1


MultiKeyGroupBy_Test().test_advanced_multikey()
MultiKeyGroupBy_Test().test_multkey()
